//
// Created by robin on 1/10/18.
//

#include <QMetaEnum>
#include "dns-base/RecordData/QDnsRecordDataA.h"
#include "dns-base/RecordData/QDnsRecordDataAAAA.h"
#include "dns-base/RecordData/QDnsRecordDataMX.h"
#include "dns-base/RecordData/QDnsRecordDataPTR.h"
#include "dns-base/RecordData/QDnsRecordDataSOA.h"
#include "dns-base/RecordData/QDnsRecordDataRaw.h"
#include "dns-base/QDnsLabel.h"
#include "dns-base/QDnsResourceRecord.h"

namespace QDns
{
    namespace Base
    {
        QDnsResourceRecord::QDnsResourceRecord()
                : m_isValid(true)
                , m_data(0)
        {
        }

        QDnsResourceRecord::QDnsResourceRecord(const QByteArray &packetData, quint32 pos)
                : QDnsResourceRecord()
        {
            QDataStream dataStream(packetData);
            dataStream.skipRawData(pos);
            parse(packetData, dataStream);
        }

        QDnsResourceRecord::QDnsResourceRecord(const QByteArray &packetData, QDataStream &dataStream)
                : QDnsResourceRecord()
        {
            parse(packetData, dataStream);
        }

        QDnsResourceRecord::QDnsResourceRecord(const QDnsResourceRecord &other)
        {
            copy(other);
        }

        QDnsResourceRecord &QDnsResourceRecord::operator=(const QDnsResourceRecord& other)
        {
            copy(other);
            return *this;
        }

        void QDnsResourceRecord::copy(const QDnsResourceRecord &other)
        {
            m_isValid = other.m_isValid;
            m_name = other.m_name;
            m_type = other.m_type;
            m_class = other.m_class;
            m_ttl = other.m_ttl;
            m_data.reset(other.getData()->clone());
        }

        void QDnsResourceRecord::parse(const QByteArray &packetData, QDataStream &dataStream)
        {
            m_name = QDnsLabel::parse(packetData, dataStream);
            if (m_name.isNull() || dataStream.device()->bytesAvailable() < 10)
            {
                m_isValid = false;
            }
            else
            {
                dataStream >> m_type;
                dataStream >> m_class;
                dataStream >> m_ttl;
                quint16 length;
                dataStream >> length;
                if (dataStream.device()->bytesAvailable() < length)
                {
                    m_isValid = false;
                }
                else
                {
                    char data[length];
                    dataStream.readRawData(data, length);
                    QByteArray byteArray(data, length);
                    QDataStream dataStream2(byteArray);
                    QDnsRecordData* recordData = nullptr;
                    if (m_type == QDns::Base::QDnsBase::RecordType::A)
                    {
                        recordData = new QDnsRecordDataA(packetData, dataStream2);
                    }
                    else if (m_type == QDns::Base::QDnsBase::RecordType::AAAA)
                    {
                        recordData = new QDnsRecordDataAAAA(packetData, dataStream2);
                    }
                    else if (m_type == QDns::Base::QDnsBase::RecordType::MX)
                    {
                        recordData = new QDnsRecordDataMX(packetData, dataStream2);
                    }
                    else if (m_type == QDns::Base::QDnsBase::RecordType::PTR)
                    {
                        recordData = new QDnsRecordDataPTR(packetData, dataStream2);
                    }
                    else if (m_type == QDns::Base::QDnsBase::RecordType::SOA)
                    {
                        recordData = new QDnsRecordDataSOA(packetData, dataStream2);
                    }
                    else
                    {
                        recordData = new QDnsRecordDataRaw(packetData, dataStream2);
                    }
                    m_data.reset(recordData);
                }
            }
        }

        QByteArray QDnsResourceRecord::serialize() const
        {
            QByteArray byteArray;
            QDataStream dataStream(&byteArray, QIODevice::WriteOnly);
            serialize(dataStream);
            return byteArray;
        }

        void QDnsResourceRecord::serialize(QDataStream &dataStream) const
        {
            auto label = QDnsLabel::serialize(m_name);
            dataStream.writeRawData(label.constData(), label.length());
            dataStream << m_type;
            dataStream << m_class;
            dataStream << m_ttl;
            QByteArray byteArray;
            QDataStream dataStream2(&byteArray, QIODevice::WriteOnly);
            m_data->serialize(dataStream2);
            dataStream << (quint16) byteArray.length();
            dataStream.writeRawData(byteArray.constData(), byteArray.length());
        }

        bool QDnsResourceRecord::isValid() const
        {
            return m_isValid;
        }

        const QString &QDnsResourceRecord::getName() const
        {
            return m_name;
        }

        void QDnsResourceRecord::setName(const QString &name)
        {
            m_name = name;
        }

        QDnsBase::RecordType QDnsResourceRecord::getType() const
        {
            return (QDnsBase::RecordType) m_type;
        }

        void QDnsResourceRecord::setType(QDnsBase::RecordType type)
        {
            m_type = type;
        }

        QDnsBase::RecordClass QDnsResourceRecord::getClass() const
        {
            return (QDnsBase::RecordClass) m_class;
        }

        void QDnsResourceRecord::setClass(QDnsBase::RecordClass aClass)
        {
            m_class = aClass;
        }

        quint32 QDnsResourceRecord::getTtl() const
        {
            return m_ttl;
        }

        void QDnsResourceRecord::setTtl(quint32 ttl)
        {
            m_ttl = ttl;
        }

        QDnsRecordData* QDnsResourceRecord::getData() const
        {
            return m_data.data();
        }

        void QDnsResourceRecord::setData(QDnsRecordData* data)
        {
            m_data.reset(data);
        }

        QString QDnsResourceRecord::toString() const
        {
            static int recordClassId = QDns::Base::QDnsBase::staticMetaObject.indexOfEnumerator("RecordClass");
            static int recordTypeId = QDns::Base::QDnsBase::staticMetaObject.indexOfEnumerator("RecordType");
            auto recordClassStr = QString(QDns::Base::QDnsBase::staticMetaObject.enumerator(recordClassId).valueToKey(getClass())).toUpper();
            auto recordTypeStr = QString(QDns::Base::QDnsBase::staticMetaObject.enumerator(recordTypeId).valueToKey(getType())).toUpper();
            return getName() + "\t" + QString::number(getTtl()) + "\t" + recordClassStr + "\t" + recordTypeStr + "\t" + getData()->toString();
        }

        QDebug operator<<(QDebug debug, const QDnsResourceRecord &rr)
        {
            QDebugStateSaver saver(debug);
            debug.noquote() << rr.toString();
            return debug;
        }
    }
}